package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

/**
 * A widget that calls the provided {@code transformer} every layout.
 */
public record CustomTransformBuilder(Widget child, Transformer transformer) implements SingleChildWidget {
	@Override
	public @NotNull AbstractState<CustomTransformBuilder> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class Transform {
		public float x;
		public float y;

		private Transform(float x, float y) {
			this.x = x;
			this.y = y;
		}
	}

	@FunctionalInterface
	public interface Transformer {
		void transform(Transform transform,
					   float computedWidth, float computedHeight,
					   float availableWidth, float availableHeight);
	}

	private static final class State extends AbstractState<CustomTransformBuilder> {
		private Transformer transformer;
		private float transformedX;
		private float transformedY;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, CustomTransformBuilder widget, boolean init) {
			super.configure(stateContext, widget, init);
			if (!widget.transformer.equals(this.transformer)) {
				this.transformer = widget.transformer;
				markNeedsScissorLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (e, ctx) -> {
				e.copyAvailableSizeFrom(this);
				e.stencilLayout(ctx);
				var transform = new Transform(e.allocatedX(), e.allocatedY());
				this.transformer.transform(transform,
						e.computedWidth(),
						e.computedHeight(),
						e.allocatedWidth(),
						e.allocatedHeight());
				boolean positionDirty = false;
				if (this.transformedX != transform.x) {
					this.transformedX = transform.x;
					positionDirty = true;
				}
				if (this.transformedY != transform.y) {
					this.transformedY = transform.y;
					positionDirty = true;
				}
				if (positionDirty) {
					markNeedsTraceLayout();
					markNeedsScissorLayout();
				}
				// TODO: is this really the ideal behaviour?
				copyComputedSizeFrom(e);
			});
		}

		@Override
		protected void computeTraceLayout() {
			var e = getChildElement();
			// TODO: should we allow overflow?
			e.setPosition(this.transformedX, this.transformedY);
			e.traceLayout();
		}

		@Override
		protected void computeScissorLayout() {
			var childElement = this.getChildElement();
			var x = Math.max(this.transformedX, 0);
			var y = Math.max(this.transformedY, 0);
			childElement.setClippedRect(x,
					y,
					childElement.computedWidth() + (this.transformedX - x),
					childElement.computedHeight() + (this.transformedY - y));
			childElement.scissorLayout();
		}
	}
}
